/*
 *  linux/fs/fatx/namei.c
 *
 *  Written 1992,1993 by Werner Almesberger
 *  Hidden files 1995 by Albert Cahalan <albert@ccs.neu.edu> <adc@coe.neu.edu>
 *  Rewritten for constant inumbers 1999 by Al Viro
 */


#define __NO_VERSION__
#include <linux/module.h>

#include <linux/sched.h>
#include <linux/fatx_fs.h>
#include <linux/errno.h>
#include <linux/string.h>

#include <asm/uaccess.h>

#define PRINTK(format, args...) do { if (fatx_debug) printk( format, ##args ); } while(0)

//#define PRINTK(format, args...) printk( format, ##args )

//static inline void do_nothing() {}
//#define PRINTK(format, args...) do_nothing()

/* Characters that are undesirable in an MS-DOS file name */
  
static char bad_chars[] = "*?<>|\";";

/*
 * Case insensitive hash of FATX name
 */
static int fatx_hashi(struct dentry *dentry, struct qstr *qstr)
{
	struct nls_table *t = MSDOS_SB(dentry->d_inode->i_sb)->nls_io;
	const char *name;
	int len;
	unsigned long hash;

	len = qstr->len;
	name = qstr->name;
	
	hash = init_name_hash();
	while (len--)
		hash = partial_name_hash(fatx_tolower(t, *name++), hash);
	qstr->hash = end_name_hash(hash);

	return 0;
}

/*
 * Case insensitive compare of two FATX names.
 */
static int fatx_cmpi(struct dentry *dentry, struct qstr *a, struct qstr *b)
{
	struct nls_table *t = MSDOS_SB(dentry->d_inode->i_sb)->nls_io;

	/* A filename cannot end in '.' or we treat it like it has none */
	if (a->len == b->len) {
		if (fatx_strnicmp(t, a->name, b->name, a->len) == 0)
			return 0;
	}
	return 1;
}

struct dentry_operations fatx_dentry_operations = {
	d_hash:		fatx_hashi,
	d_compare:	fatx_cmpi,
};

/***** Formats a FATX file name. Rejects invalid names. */
static int fatx_format_name( const char *name,int len,char *out_name )
{
	int i;
	char trash[FATX_MAX_NAME_LENGTH];

	if (len > FATX_MAX_NAME_LENGTH) return -EINVAL;
	
	if (out_name == NULL) out_name = trash;
	
	//check for bad characters in name
	for(i=0; i<len; i++) {
		if (strchr(bad_chars,name[i])) return -EINVAL;
		out_name[i] = name[i];
	}
	for(;i<FATX_MAX_NAME_LENGTH;i++) {
		out_name[i] = 0xFF;
	}
	return 0;
}

/***** Locates a directory entry.  Uses unformatted name. */
static int fatx_find(
		struct inode *dir,
		const char *name,
		int len,
		struct buffer_head **bh,
		struct fatx_dir_entry **de,
		int *ino )
{
	//verify its a valid name
	if (fatx_format_name(name,len,NULL) < 0) return -ENOENT;
	
	//find the name in the directory
	return fatx_scan(dir,name,len,bh,de,ino);

}

/***** Get inode using directory and name */
struct dentry *fatx_lookup(struct inode *dir,struct dentry *dentry)
{
	struct super_block *sb = dir->i_sb;
	struct inode *inode = NULL;
	struct fatx_dir_entry *de;
	struct buffer_head *bh = NULL;
	int ino,res;
	
	PRINTK("fatx_lookup\n");

	dentry->d_op = &fatx_dentry_operations;

	res = fatx_find(dir, dentry->d_name.name, dentry->d_name.len, &bh,
			&de, &ino);

	if (res == -ENOENT)
		goto add;
	if (res < 0)
		goto out;
	inode = fatx_build_inode(sb, de, ino, &res);
	if (res)
		goto out;
add:
	d_add(dentry, inode);
	res = 0;
out:
	if (bh)
		fat_brelse(sb, bh);
	return ERR_PTR(res);
}

/***** Creates a directory entry (name is already formatted). */
static int fatx_add_entry(
		struct inode *dir, 
		const char *name,
		int len,
		struct buffer_head **bh,
		struct fatx_dir_entry **de,
		int *ino,
		int is_dir )
{
	struct super_block *sb = dir->i_sb;
	int res;

	if ((res = fatx_do_add_entry(dir, bh, de, ino))<0)
		return res;
	/*
	 * XXX all times should be set by caller upon successful completion.
	 */
	dir->i_ctime = dir->i_mtime = CURRENT_TIME;
	mark_inode_dirty(dir);
	memcpy((*de)->name,name,len);
	(*de)->name_length = len;
	(*de)->attr = is_dir ? ATTR_DIR : ATTR_ARCH;
	(*de)->start = 0;
	fatx_date_unix2dos(dir->i_mtime,&(*de)->time,&(*de)->date);
	(*de)->size = 0;
	fat_mark_buffer_dirty(sb, *bh);
	return 0;
}

/***** Create a file */
int fatx_create(struct inode *dir,struct dentry *dentry,int mode)
{
	struct super_block *sb = dir->i_sb;
	struct buffer_head *bh;
	struct fatx_dir_entry *de;
	struct inode *inode;
	int ino,res;
	const char *name = dentry->d_name.name;
	int name_length = dentry->d_name.len;

	res = fatx_format_name(name,name_length,NULL);
	if (res < 0)
		return res;
	
	if (fatx_scan(dir,name,name_length,&bh,&de,&ino) >= 0) {
		fat_brelse(sb, bh);
		return -EINVAL;
 	}
	inode = NULL;
	res = fatx_add_entry(dir, name, name_length, &bh, &de, &ino, 0);
	if (res)
		return res;
	inode = fatx_build_inode(dir->i_sb, de, ino, &res);
	fat_brelse(sb, bh);
	if (!inode)
		return res;
	inode->i_mtime = inode->i_atime = inode->i_ctime = CURRENT_TIME;
	mark_inode_dirty(inode);
	d_instantiate(dentry, inode);
	return 0;
}

/***** Remove a directory */
int fatx_rmdir(struct inode *dir, struct dentry *dentry)
{
	struct super_block *sb = dir->i_sb;
	struct inode *inode = dentry->d_inode;
	int res,ino;
	struct buffer_head *bh;
	struct fatx_dir_entry *de;

	bh = NULL;
	res = fatx_find(dir, dentry->d_name.name, dentry->d_name.len,
				&bh, &de, &ino);
	if (res < 0)
		goto rmdir_done;
	/*
	 * Check whether the directory is not in use, then check
	 * whether it is empty.
	 */
	res = fatx_dir_empty(inode);
	if (res)
		goto rmdir_done;

	de->name_length = DELETED_FLAG;
	fat_mark_buffer_dirty(sb, bh);
	fat_detach(inode);
	inode->i_nlink = 0;
	inode->i_ctime = dir->i_ctime = dir->i_mtime = CURRENT_TIME;
	dir->i_nlink--;
	mark_inode_dirty(inode);
	mark_inode_dirty(dir);
	res = 0;

rmdir_done:
	fat_brelse(sb, bh);
	return res;
}

/***** Make a directory */
int fatx_mkdir(struct inode *dir,struct dentry *dentry,int mode)
{
	struct super_block *sb = dir->i_sb;
	struct buffer_head *bh;
	struct fatx_dir_entry *de;
	struct inode *inode;
	int res;
	const char *name = dentry->d_name.name;
	int name_length = dentry->d_name.len;
	int ino;

	res = fatx_format_name(name,name_length,NULL);
	if (res < 0)
		return res;
	if (fatx_scan(dir,name,name_length,&bh,&de,&ino) >= 0)
		goto out_exist;

	res = fatx_add_entry(dir, name, name_length, &bh, &de, &ino, 1);
	if (res)
		goto out_unlock;
	inode = fatx_build_inode(dir->i_sb, de, ino, &res);
	if (!inode) {
		fat_brelse(sb, bh);
		goto out_unlock;
	}
	res = 0;

	dir->i_nlink++;
	inode->i_nlink = 2; /* no need to mark them dirty */

	res = fatx_new_dir(inode, dir);
	if (res)
		goto mkdir_error;

	fat_brelse(sb, bh);
	d_instantiate(dentry, inode);
	res = 0;

out_unlock:
	return res;

mkdir_error:
	printk(KERN_WARNING "fatx_mkdir: error=%d, attempting cleanup\n", res);
	inode->i_nlink = 0;
	inode->i_ctime = dir->i_ctime = dir->i_mtime = CURRENT_TIME;
	dir->i_nlink--;
	mark_inode_dirty(inode);
	mark_inode_dirty(dir);
	de->name_length = DELETED_FLAG;
	fat_mark_buffer_dirty(sb, bh);
	fat_brelse(sb, bh);
	fat_detach(inode);
	iput(inode);
	goto out_unlock;

out_exist:
	fat_brelse(sb, bh);
	res = -EINVAL;
	goto out_unlock;
}

/***** Unlink a file */
int fatx_unlink( struct inode *dir, struct dentry *dentry)
{
	struct super_block *sb = dir->i_sb;
	struct inode *inode = dentry->d_inode;
	int res,ino;
	struct buffer_head *bh;
	struct fatx_dir_entry *de;

	bh = NULL;
	res = fatx_find(dir, dentry->d_name.name, dentry->d_name.len,
			&bh, &de, &ino);
	if (res < 0)
		goto unlink_done;

	de->name_length = DELETED_FLAG;
	fat_mark_buffer_dirty(sb, bh);
	fat_detach(inode);
	fat_brelse(sb, bh);
	inode->i_nlink = 0;
	inode->i_ctime = dir->i_ctime = dir->i_mtime = CURRENT_TIME;
	mark_inode_dirty(inode);
	mark_inode_dirty(dir);
	res = 0;
unlink_done:
	return res;
}

static int do_fatx_rename(
		struct inode *old_dir, 
		struct dentry *old_dentry,
		struct inode *new_dir,
		struct dentry *new_dentry,
		struct buffer_head *old_bh,
		struct fatx_dir_entry *old_de,
		int old_ino )
{
	struct super_block *sb = old_dir->i_sb;
	struct buffer_head *new_bh=NULL,*dotdot_bh=NULL;
	struct fatx_dir_entry *new_de,*dotdot_de;
	struct inode *old_inode,*new_inode;
	int new_ino,dotdot_ino;
	int error;
	int is_dir;
	const char *new_name = new_dentry->d_name.name;
	int new_name_len = old_dentry->d_name.len;

	PRINTK("FATX: do_fatx_rename: entered\n");
	
	old_inode = old_dentry->d_inode;
	new_inode = new_dentry->d_inode;
	is_dir = S_ISDIR(old_inode->i_mode);

	error = fatx_scan(new_dir,new_name,new_name_len,&new_bh,&new_de,&new_ino);
	if (error>=0 &&!new_inode)
		goto degenerate_case;
	if (is_dir) {
		if (new_inode) {
			error = fatx_dir_empty(new_inode);
			if (error)
				goto out;
		}
		
		//FIXME: this is wrong...no such dir name...
		error = fatx_scan(old_inode, "..", 2 , &dotdot_bh,
				&dotdot_de, &dotdot_ino);
		if (error < 0) {
			printk(KERN_WARNING
				"FATX: %s/%s, get dotdot failed, ret=%d\n",
				old_dentry->d_parent->d_name.name,
				old_dentry->d_name.name, error);
			goto out;
		}
	}
	if (!new_bh) {
		error = fatx_add_entry(
				new_dir, 
				new_name, 
				new_name_len, 
				&new_bh, 
				&new_de,
				&new_ino, 
				is_dir );
		if (error)
			goto out;
	}
	new_dir->i_version = ++event;

	/* There we go */

	if (new_inode)
		fat_detach(new_inode);
	old_de->name_length = DELETED_FLAG;
	fat_mark_buffer_dirty(sb, old_bh);
	fat_detach(old_inode);
	fat_attach(old_inode, new_ino);
	MSDOS_I(old_inode)->i_attrs &= ~ATTR_HIDDEN;
	mark_inode_dirty(old_inode);
	old_dir->i_version = ++event;
	old_dir->i_ctime = old_dir->i_mtime = CURRENT_TIME;
	mark_inode_dirty(old_dir);
	if (new_inode) {
		new_inode->i_nlink--;
		new_inode->i_ctime = CURRENT_TIME;
		mark_inode_dirty(new_inode);
	}
	if (dotdot_bh) {
		dotdot_de->start = CT_LE_L(MSDOS_I(new_dir)->i_logstart);
		
		PRINTK("FATX: do_fatx_rename: start = %08lX (LE=%08lX)\n",
				(long)MSDOS_I(new_dir)->i_logstart,
				(long)dotdot_de->start);
		
		fat_mark_buffer_dirty(sb, dotdot_bh);
		old_dir->i_nlink--;
		mark_inode_dirty(old_dir);
		if (new_inode) {
			new_inode->i_nlink--;
			mark_inode_dirty(new_inode);
		} else {
			new_dir->i_nlink++;
			mark_inode_dirty(new_dir);
		}
	}
	error = 0;
out:
	fat_brelse(sb, new_bh);
	fat_brelse(sb, dotdot_bh);
	PRINTK("FATX: do_fatx_rename: leaving (normal)\n");
	return error;

degenerate_case:
	error = -EINVAL;
	if (new_de!=old_de)
		goto out;
	MSDOS_I(old_inode)->i_attrs &= ~ATTR_HIDDEN;
	mark_inode_dirty(old_inode);
	old_dir->i_version = ++event;
	old_dir->i_ctime = old_dir->i_mtime = CURRENT_TIME;
	mark_inode_dirty(old_dir);
	PRINTK("FATX: do_fatx_rename: leaving (degenerate)\n");
	return 0;
}

/***** Rename, a wrapper for rename_same_dir & rename_diff_dir */
int fatx_rename(struct inode *old_dir,struct dentry *old_dentry,
		 struct inode *new_dir,struct dentry *new_dentry)
{
	struct super_block *sb = old_dir->i_sb;
	struct buffer_head *old_bh;
	struct fatx_dir_entry *old_de;
	int old_ino, error;

	error = fatx_format_name(old_dentry->d_name.name,old_dentry->d_name.len,NULL);
	if (error < 0)
		goto rename_done;
	error = fatx_format_name(new_dentry->d_name.name,new_dentry->d_name.len,NULL);
	if (error < 0)
		goto rename_done;

	error = fatx_scan(
			old_dir, 
			old_dentry->d_name.name, 
			old_dentry->d_name.len,
			&old_bh, 
			&old_de, 
			&old_ino );
	if (error < 0)
		goto rename_done;

	error = do_fatx_rename(
			old_dir, 
			old_dentry,
			new_dir, 
			new_dentry,
			old_bh, 
			old_de, 
			(ino_t)old_ino );
	fat_brelse(sb, old_bh);

rename_done:
	return error;
}
